/*
 @licstart  The following is the entire license notice for the JavaScript code in this page.

 Copyright (C) 2014-2019, Josef Hahn

 This file is part of PyWebApp.

 PyWebApp is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 PyWebApp is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with PyWebApp.  If not, see <http://www.gnu.org/licenses/>.

 @licend  The above is the entire license notice for the JavaScript code in this page.
*/

pywebapp = new Object();
(function(){

    $(document).ready(function(){ pywebapp.hasExternalCloseControls = Boolean(window.pyhtmlview); });

    window.onerror = function(e, src, line) {
        if (!pywebapp._processesevents)
            return;
        pywebapp.ajax({
            url: 'unhandled_client_error',
            data: { error: String(e), src: String(src||""), line: String(line||"") },
            method: 'POST',
            onsuccess: function(data) {
                if (data && data.shutdown)
                    pywebapp.shutdown();
            }
        });
        return true;
    };

    clove.i18n.addString('PyWebAppIsShutDown', {en: "This application was stopped.\n\nYou can close this browser tab now.", de: "Diese Anwendung wurde beendet.\n\nSie können dieses Browser-Tab nun schließen."});
    clove.i18n.addString('PyWebAppClosedNotificationText', {en: "It seems you have closed the browser tab which shows this application.\n\nIf you did that with the intention to close the application, you can do that right now, or simply continue.", de: "Scheinbar haben Sie das Browser-Tab geschlossen, das diese Anwendung angezeigt hat.\n\nWenn Sie den Wunsch hatten, die Anwendung zu schließen, können Sie das nun direkt tun, oder einfach fortfahren."});
    clove.i18n.addString('PyWebAppPyHtmlViewSupportConfirmClose', {en: "Do you really want to close the application?", de: "Möchten Sie die Anwendung wirklich schließen?"});
    clove.i18n.addString('PyWebAppClosedNotificationButtonLabel', {en: "Close the application", de: "Diese Anwendung schließen"});
    clove.i18n.addString('PyWebAppProgramErrorOccurred', {en: "An internal program error occurred!\n\nTo be safe, you should stop and restart the application now.", de: "Es ist ein interner Programmfehler aufgetreten.\n\nUm sicherzugehen sollten Sie die Anwendung beenden und neu starten."});
    clove.i18n.addString('PyWebAppShowDetails', {en: "Show details", de: "Zeige Details"});
    clove.i18n.addString('PyWebAppHideDetails', {en: "Hide details", de: "Verstecke Details"});
    clove.i18n.addString('PyWebAppOpenDir', {en: "Open Directory", de: "Verzeichnis öffnen"});
    clove.i18n.addString('PyWebAppOpenFile', {en: "Open File", de: "Datei öffnen"});
    clove.i18n.addString('PyWebAppGoto', {en: "Go to", de: "Gehe zu"});
    clove.i18n.addString('PyWebAppCancel', {en: "Cancel", de: "Abbrechen"});
    clove.i18n.addString('PyWebAppCurrent', {en: "Current", de: "Aktuell"});
    clove.i18n.addString('PyWebAppFilesystem', {en: "Filesystem", de: "Dateisystem"});
    clove.i18n.addString('PyWebAppDrives', {en: "Drives", de: "Laufw."});

    function _errorpopup(errortext) {
        clove.notifications.notify({
            rows: [
                {label: clove.i18n.strPyWebAppProgramErrorOccurred},
                {
                    view: 'Expander', expandedLabel: clove.i18n.strPyWebAppHideDetails,
                    collapsedLabel: clove.i18n.strPyWebAppShowDetails,
                    body: {
                       label: errortext,
                       styleClass: 'pywebapp_errornotification_details',
                       horizontalStretchAffinity: 1,
                    },
                },
            ],
        }, {
            styleClass: 'pywebapp_errornotification',
            closeable: true,
        });
    }

    var _eventhandlers = new Object();

    pywebapp.addEventHandler = function(key, fct) {
        if (!_eventhandlers.hasOwnProperty(key))
            _eventhandlers[key] = [];
        _eventhandlers[key].push(fct);
    }

    var _busytokens = [];

    function _processEvent(event) {
        if (event.name == "stopapplication") {
            pywebapp.isrunning = false;
            pywebapp.uiShutdown();
            if (event.targeturl)
                location.href = event.targeturl;
        }
        else if (event.name == "showunhandlederror")
            _errorpopup(event.errortext);
        else if (event.name == "DataStore_update") {
            var ds = pywebapp.getDatasource(event.datasource);
            var objids = event.id;
            var bo = ds._rootnode;
            for (var i=0; i<objids.length; i++) {
                var objid = objids[i];
                var rcell;
                for (var iro=0; iro<bo.childs.length; iro++) {
                    var row = bo.childs[iro];
                    for (var ice=0; row.length; ice++) {
                        var cell = row[ice];
                        if (cell == objid) {
                            rcell = cell;
                            break;
                        }
                    }
                    if (rcell)
                        break;
                }
                if (!rcell)
                    return;
                else
                    bo = rcell;
            }
            ds.async_pull(bo.valuePointer(), {reason: 'PyWebApp_DataStore', sub: 'update'});
        }
        var customhandlers = _eventhandlers[event.name];
        if (customhandlers)
            for(var ic=0; ic<customhandlers.length; ic++)
                customhandlers[ic](event);
    }

    function _getNextEvent() {
        pywebapp._processesevents = true;
        if (!pywebapp.isrunning)
            return;
        pywebapp.ajax({
            url: 'pullevent',
            data: {lastid: pywebapp.eventlastid},
            method: 'GET',
            onsuccess: function(data) {
                for(var i=0; i<data.length; i++) {
                    var ddata = data[i];
                    pywebapp.eventlastid = Math.max(pywebapp.eventlastid, ddata.id);
                    if (ddata.data)
                        try {
                            _processEvent(ddata.data);                        
                        } catch (e) {
                            setTimeout(function(){throw e;});
                        }
                }
                setTimeout(_getNextEvent, 0);
            },
            onerror: function() {
                pywebapp.uiShutdown();
            }
        });
    }
    
    pywebapp.startEventProcessing = _getNextEvent;

    class AjaxAsyncDatasource extends clove.AjaxAsyncDatasource {

        constructor(name) {
            super({
                data_pullUrl: "datastore_pull",
                data_pushUrl: "datastore_push",
                params: {name: name},
                rootType: "data",
                ajaxAdditionalArgs: {dataType: 'json'},
            });
            this.name = name;
        }
        
    }
    
    var _datasources = new Object();
    
    pywebapp.getDatasource = function(name) {
        if (!_datasources.hasOwnProperty(name))
            _datasources[name] = new AjaxAsyncDatasource(name);
        return _datasources[name];
    }
    
    var _head1;
    var _head2;
    
    function _setwindowtitle() {
        var title = (_head1!==undefined) ? _head1 : "";
        if (_head2!==undefined)
            title += ": " + _head2;
        document.title = title;
    }

    class Configuration {
     
        constructor() {
            this._handlers = {};
            this._appliedvalues = undefined;
            pywebapp.addEventHandler("AppConfig_setconfig", function(e){
                pywebapp.configuration.applyValue(e.key, e.value);
            });
        }

        /*!
        * This function is called only internally.
        */
        applyValues(dict) {
            this._appliedvalues = {};
            for(var key in dict)
                if (dict.hasOwnProperty(key))
                    this._appliedvalues[key] = dict[key];
        }

        /*!
        * This function is called only internally.
        */
        applyValue(key, value) {
            var hasappliedvalue = this._appliedvalues.hasOwnProperty(key);
            var oldappliedvalue = this._appliedvalues[key];
            var isnewvalue = (!hasappliedvalue || (oldappliedvalue !== value));
            this._appliedvalues[key] = value;
            var handlers = this._handlers[key];
            if (handlers)
                for (var ih=0;ih<handlers.length;ih++) {
                    var hh= handlers[ih];
                    if (isnewvalue || hh[1])
                        hh[0](value, oldappliedvalue, hasappliedvalue);
                }
        }
        
        addHandler(key, fct, runAlsoOnNeutralAssignments) {
            if (this._appliedvalues === undefined)
                throw new Error("pywebapp.configuration.addHandler called before pywebapp was ready.");
            if (!this._handlers.hasOwnProperty(key))
                this._handlers[key] = [];
            this._handlers[key].push([fct, runAlsoOnNeutralAssignments]);
            if (this._appliedvalues.hasOwnProperty(key)) {
                var value = this._appliedvalues[key];
                fct(value, undefined, false);
            }
        }
        
        getValue(key) {
            return this._appliedvalues[key];
        }

        setValue(key, value) {
            pywebapp.ajax({
                url: 'setappconfigvalue',
                data: {key: key, value: JSON.stringify(value)},
            });
        }
        
    }
    
    pywebapp.configuration = new Configuration();
    
    class AbstractFilesystemDialog {

        constructor(config) {
            this.config = clove.utils.applyDefaultsToConfig(config, {
                startPath: "/",
                title: clove.i18n.strPyWebAppFilesystem,
                headerWidget: {height: '1px'},
                headerText: "",
                showHidden: true,
                onlyDirs: false,
                onsuccess: function(){},
                oncancel: function(){},
            });
        }

        _listfsdata(path) {
            return {path: path, showHidden: this.config.showHidden?"1":"0", onlyDirs: this.config.onlyDirs?"1":"0"};
        }

        _showdir(path) {
            var self = this;
            this.dirlist.setDatasource();
            this.lblcurrdir.setLabel(clove.i18n.strPyWebAppCurrent + ": ...");
            this.btngoto.setEnabled(false);
            pywebapp.ajax({
                url: 'listfs',
                data: this._listfsdata(path),
                onsuccess: function(data) {
                    self.btngoto.setEnabled(true);
                    self.currpath = data.path;
                    self.lblcurrdir.setLabel(clove.i18n.strPyWebAppCurrent + ": " + data.path);
                    var ds = clove.utils.arrayToDatasource(data.dirs);
                    ds.insertColumn(0);
                    for(var ir=0;ir<ds.rowCount();ir++) {
                        var ptr = ds.valuePointer(ir, 0);
                        ds.setValue(ptr, data.dirs[ir][1]);
                        ds.setMetadata(ptr, "icon", clove.Icon.bySymbol(data.dirs[ir][0] ? "⠀" : "📂"));
                    }
                    self.dirlist.setDatasource(ds);
                },
            });
        }

        showDialog() {
            var self = this;
            var dlgres = clove.Dialog.show({
                view: 'clove.Dialog',
                title: this.config.title,
                body: {
                    rows: [
                        {label: this.config.headerText, styleClass: 'clovestyle_vmargin_tiny clovestyle_hmargin_small', width: '1px', horizontalStretchAffinity: 1},
                        this.config.headerWidget,
                        {rows: [{label: '', name: 'lblcurrdir', styleClass: 'pywebapp_filebrowser_currentdir'}], styleClass: 'clovestyle_panel_bar'},
                        {view: 'ScrollView', height: '150pt', horizontalStretchAffinity: 1, body:
                            {view: 'DataView', showOnlyFirstColumn: true, name: 'dirlist', allowSelection: true, hideExpanders: true, gridVisible: false, granularity: 'row', styleClass: 'clovestyle_margin_small'},
                        },
                        {cols: [
                            {view: 'Button', label: clove.i18n.strPyWebAppGoto, name: 'btngoto', styleClass: 'clovestyle_marginnotleft_small', icon: clove.Icon.bySymbol(clove.symbols.CheckMark)},
                            {view: 'Button', label: clove.i18n.strPyWebAppDrives, name: 'btnchangedrive', styleClass: 'clovestyle_marginnotleft_small'},
                            {view: 'Button', label: clove.i18n.strPyWebAppCancel, name: 'btncancel', styleClass: 'clovestyle_marginnotleft_small'},
                        ]},
                    ],
                    width: '330pt',
                },
            }, {
                modal: true,
            });
            this.dirlist = dlgres.namescope.getByName('dirlist');
            this.lblcurrdir = dlgres.namescope.getByName('lblcurrdir');
            this.btncancel = dlgres.namescope.getByName('btncancel');
            this.btngoto = dlgres.namescope.getByName('btngoto');
            this.btnchangedrive = dlgres.namescope.getByName('btnchangedrive');
            this.currpath = undefined;
            if (!pywebapp.configuration.getValue("AbstractFilesystemDialog_showChangeDriveButton"))
                this.btnchangedrive.setVisibility(clove.InvisibleCollapsed);
            this.dirlist.OnNodeActivated.addHandler(function(e){
                var eds = e.node.datasource;
                var npath = self.currpath + (self.currpath=="/" ? "" : "/") + eds.getValue(e.node);
                var isfile = eds.getValue(eds.valuePointer(e.node.irow, 1))[0];
                if (isfile) {
                    dlgres.widget.close();
                    if (self.config.onsuccess)
                        self.config.onsuccess(npath);
                }
                else
                    self._showdir(npath);
            });
            this.btncancel.OnClicked.addHandler(function(e){
                dlgres.widget.close();
                if (self.config.oncancel)
                    self.config.oncancel();
            });
            this.btngoto.OnClicked.addHandler(function(e){
                dlgres.widget.close();
                if (self.config.onsuccess)
                    self.config.onsuccess(self.currpath);
            });
            this.btnchangedrive.OnClicked.addHandler(function(e){
                self._showdir("");
            });
            this._showdir(self.config.startPath);
            return dlgres;
        }

    }

    pywebapp.AbstractFilesystemDialog = AbstractFilesystemDialog;

    class OpenFileDialog extends pywebapp.AbstractFilesystemDialog{

        constructor(config) {
            super(clove.utils.applyDefaultsToConfig(config, {
                title: clove.i18n.strPyWebAppOpenFile,
            }));
        }

        _showdir(path) {
            this.btngoto.setVisibility(clove.InvisibleCollapsed);
            super._showdir(path);
        }

    }

    pywebapp.OpenFileDialog = OpenFileDialog;

    class OpenDirectoryDialog extends pywebapp.AbstractFilesystemDialog{

        constructor(config) {
            super(clove.utils.applyDefaultsToConfig(config, {
                title: clove.i18n.strPyWebAppOpenDir,
                onlyDirs: true,
            }));
        }

    }

    pywebapp.OpenDirectoryDialog = OpenDirectoryDialog;

    class UserFeedback {

        constructor() {
            this._currentrequest = undefined;
            this._currentrequestdlg = undefined;
        }

        initialize() {
            var self = this;
            this._handlers = {};
            this.registerHandler("message", function(feedbackcfg){
                var result = clove.messageDialog({message: feedbackcfg.message, buttons: feedbackcfg.buttons});
                result.addHandler(function(e){
                    self.answerRequest(feedbackcfg, e.button);
                });
                return result.dlgres.widget;
            });
            this.registerHandler("input", function(feedbackcfg){
                var result = clove.inputDialog({question: feedbackcfg.question, defaultAnswer: feedbackcfg.defaultanswer});
                result.addHandler(function(e){
                    self.answerRequest(feedbackcfg, e.answer);
                });
                return result.dlgres.widget;
            });
            this.registerHandler("multilineinput", function(feedbackcfg){
                var result = clove.inputDialog({question: feedbackcfg.question, defaultAnswer: feedbackcfg.defaultanswer,
                                        contentWidgetConfig: {view: 'MultilineEditBox', width: '450pt', height: '400pt',
                                                OnKeyDown: function(e) {
                                                    if (e.keyCode == 9) {
                                                        var v = e.widget.text();
                                                        var cursorPos = $(e.widget.inputnode).prop('selectionStart');
                                                        e.widget.setText(v.substring(0, cursorPos) + "    " + v.substring(cursorPos));
                                                        $(e.widget.inputnode).prop('selectionStart', cursorPos + 4);
                                                        $(e.widget.inputnode).prop('selectionEnd', cursorPos + 4);
                                                        e.originalEvent.stopPropagation();
                                                        e.originalEvent.preventDefault();
                                                    }
                                                }},
                                        withKeyboardControl: false});
                result.addHandler(function(e){
                    self.answerRequest(feedbackcfg, e.answer);
                });
                return result.dlgres.widget;
            });
            this.registerHandler("filesystem", function(feedbackcfg){
                var clss = (feedbackcfg.fstype == "directory") ? pywebapp.OpenDirectoryDialog : pywebapp.OpenFileDialog;
                return (new clss({
                    onsuccess: function(newpath) {
                        self.answerRequest(feedbackcfg, newpath);
                    },
                    oncancel: function() {
                        self.answerRequest(feedbackcfg, undefined);
                    },
                    headerText: feedbackcfg.question,
                    startPath: feedbackcfg.startpath,
                })).showDialog().widget;
            });
            setTimeout(function(){  // do that somewhat later so others have a chance to register handlers
                pywebapp.configuration.addHandler("UserFeedback_currentfeedback", function(newv, oldv, existsv) {
                    if (!self._currentrequest || !newv || self._currentrequest.reqid != newv.reqid) {
                        if (self._currentrequestdlg) {
                            self._currentrequestdlg.remove();
                            self._currentrequestdlg = undefined;
                        }
                        self._currentrequest = newv;
                        if (newv) {
                            self._currentrequestdlg = self._handlers[newv.kind](newv);
                            if (!self._currentrequestdlg)
                                throw new Error("It is not allowed for a PyWebApp UserFeedback handler to return nothing.");
                        }
                    }
                });
            });
        }

        answerRequest(feedbackcfg, answer) {
            pywebapp.ajax({
                url: 'answer_userfeedback',
                data: {reqid: feedbackcfg.reqid, answer: JSON.stringify([answer])},
                method: 'POST',
            });
        }

        registerHandler(kind, fct) {
            this._handlers[kind] = fct;
        }

    }

    pywebapp.userfeedback = new UserFeedback();

    class PyHtmlViewSupport {

        _onwindowclose() {
            if (!pywebapp.isrunning)
                return true;
            else if (pywebapp.configuration.getValue("stopImplicitlyWhenBrowserClosed")) {
                pywebapp.shutdown();
                return true;
            }
            else if (pywebapp.configuration.getValue("showBrowserClosedNotification")) {
                var dlg = clove.messageDialog({message: clove.i18n.strPyWebAppPyHtmlViewSupportConfirmClose,
                                               buttons: [clove.i18n.strYes, clove.i18n.strNo]});
                dlg.addHandler(function(e){
                    if (e.button == 0) {
                        pywebapp.OnUiShutdown.addHandler(function(){
                            $.ajax({
                                url: 'application_tryclosebrowser',
                                method: 'POST',
                                data: {checks: '0'},
                                dataType: 'json',
                            });
                        });
                        pywebapp.shutdown({}, true);
                    }
                });
            }
        }

    }

    pywebapp.pyhtmlviewsupport = new PyHtmlViewSupport();

    var _display_shutdown_run = false;

    pywebapp.OnUiShutdown = new clove.Event();
    pywebapp.OnUiShutdown.addHandler(function(){
        if (!pywebapp.configuration.getValue("skipShutdownDialog")) {
            clove.Dialog.show({
                view: 'Dialog',
                title: pywebapp.configuration.getValue("head1"),
                body: {cols: [
                    {view: 'IconView', icon: clove.Icon.bySymbol(clove.symbols.Warning), styleClass: 'clovestyle_margin_medium'},
                    {label: clove.i18n.strPyWebAppIsShutDown, styleClass: 'clovestyle_margin_medium'},
                ]}
            }, {
                modal: true,
            });
        }
    });

    pywebapp.uiShutdown = function () {
        if (_display_shutdown_run)
            return;
        _display_shutdown_run = true;
        pywebapp.OnUiShutdown.trigger();
    }

    var _isready = false;
    var _ready_handlers = [];

    pywebapp.ready = function(fct) {
        if (_isready)
            setTimeout(fct);
        else
            _ready_handlers.push(fct);
    }

    pywebapp.initialize = function(isrunning, reopened, eventlastid, appconfigs) {
        pywebapp.isrunning = isrunning;
        pywebapp.eventlastid = eventlastid;
        $(document).ready(function(){
            clove.populateUI(function(){
                if (isrunning) {
                    clove.build({view: 'MainView', name: 'mainview'});
                    pywebapp.mainview = clove.getByName('mainview');
                    pywebapp._parseajaxdata(appconfigs);
                    pywebapp.configuration.applyValues(appconfigs);
                    pywebapp.configuration.addHandler("takeoverurl", function(s) {
                        if (s)
                            location.href = s;
                    });
                    pywebapp.configuration.addHandler("head1", function(s) {
                        _head1 = s;
                        pywebapp.mainview.setHead1(s);
                        if (!pywebapp.configuration.getValue("skipsetwindowtitle"))
                            _setwindowtitle();
                    });
                    pywebapp.configuration.addHandler("head2", function(s) {
                        _head2 = s;
                        pywebapp.mainview.setHead2(s);
                        if (!pywebapp.configuration.getValue("skipsetwindowtitle"))
                            _setwindowtitle();
                    });
                    pywebapp.configuration.addHandler("icon", function(src) {
                        if (src)
                            pywebapp.mainview.setIcon((src.length==1) ? clove.Icon.bySymbol(src) : clove.Icon.byUrl(src));
                        else
                            pywebapp.mainview.setIcon(undefined);
                    });
                    pywebapp.userfeedback.initialize();
                    _isready = true;
                    for(var i=0; i<_ready_handlers.length; i++)
                        _ready_handlers[i]();
                    _ready_handlers = undefined;
                    if (!pywebapp.configuration.getValue("isStaticApplication"))
                        pywebapp.startEventProcessing();
                    if (pywebapp.configuration.getValue('showBrowserClosedNotification')) {
                        if (reopened) {
                            var notif = clove.notifications.notify({
                                rows: [
                                    {label: clove.i18n.strPyWebAppClosedNotificationText},
                                    {cols: [{view: 'Button', label: clove.i18n.strPyWebAppClosedNotificationButtonLabel, name: 'btnclose'}]},
                                ]
                            },{
                                timeout: 30,
                                closeable: true,
                            });
                            notif.getByName('btnclose').OnClicked.addHandler(function(){
                                pywebapp.shutdown();
                            });
                        }
                    }
                }
                else {
                    pywebapp.uiShutdown();
                }
            });
        });
    }

    pywebapp.shutdown = function(data, withShutdownDialog) {
        if (!pywebapp.configuration.getValue("skipShutdownDialog") || withShutdownDialog)
            var closeanimation = clove.utils.popup({view: 'Widget'}, {modal: true, fullyTransparent: false}).widget;
        else
            var closeanimation = undefined;
        $.ajax({
            url: 'application_shutdown',
            method: 'POST',
            dataType: 'json',
            data: data,
            success: function() {
                if (closeanimation)
                    closeanimation.remove();
                pywebapp.uiShutdown();
            }
        });
    }

    pywebapp._parseajaxdata = function(data, _p, _n) {
        if (data && typeof data !== "string") {
            for(var foo in data)
                pywebapp._parseajaxdata(data[foo], data, foo);
            if (data._pywebapp_clientstring === true)
                _p[_n] = new pywebapp.ClientString(data.stringname, data.args);
        }
    }

    clove._ajaxprovider = function(config) {
        if (config.success) {
            var isuccessfct = config.success;
            config.success = function(data) {
                pywebapp._parseajaxdata(data);
                isuccessfct(data);
            }
        }
        $.ajax(config);
    }
    
    pywebapp.ajax = function(config) {
        var configurl;
        config = clove.utils.applyDefaultsToConfig(config, {
            url: 'get',
            method: 'GET',
            dataType: 'json',
            onfinished: function(){},
            onsuccess: function(){},
            onerror: function(){if (pywebapp.isrunning) throw new Error("low-level backend error in "+configurl);},
            alsoAfterShutdown: false,
        });
        configurl = config.url;
        if (!config.alsoAfterShutdown && !pywebapp.isrunning)
            return;
        config.success = function(data) {
            if (!config.alsoAfterShutdown && !pywebapp.isrunning)
                return;
            if (config.onsuccess)
                config.onsuccess(data);
            if (config.onfinished)
                config.onfinished(data);
        };
        config.error = function() {
            if (!config.alsoAfterShutdown && !pywebapp.isrunning)
                return;
            if (config.onerror)
                config.onerror();
            if (config.onfinished)
                config.onfinished();
        };
        clove._ajaxprovider(config);
    }

    class ClientString {

        constructor(stringname, args) {
            this.stringname = stringname;
            this.args = args;
        }

        toString() {
            var result = clove.i18n['str' + this.stringname];
            var replacements = [];
            for(var keyname in this.args) {
                var keystring = "{" + keyname + "}";
                var keyvalue = this.args[keyname];
                var cur = 0;
                while (cur < result.length) {
                    var i = result.indexOf(keystring, cur);
                    if (i == -1)
                        break;
                    replacements.push([i, i+keystring.length, keyvalue]);
                    cur = i + keystring.length;
                }
            }
            replacements.sort(function (a, b) { return b[0] - a[0]; });
            for(let replacement of replacements)
                result = result.slice(0, replacement[0]) + replacement[2] + result.slice(replacement[1])
            return result;
        }

    }

    pywebapp.ClientString = ClientString;

})();
